using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Reflection;
using System.Diagnostics;
using System.Threading;

namespace CPIDR
{
  public partial class Form1 : Form
  {

    #region Creation and initialization of the form and its tools

    public Form1()
    {
      InitializeComponent();
      this.Text = VersionString();
      InitializeTextBoxes();
    }

    private void Form1_Shown(object sender, EventArgs e)
    {
      this.Update();
      InitializeTagger();   // do this as soon as Form 1 is completely visible the first time
    }

    public static string VersionString()
    {
      return
        "CPIDR " +
        System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString();
    }

    void InitializeTextBoxes()
    {
      textBoxResults.Clear();
      textBoxResults.Text = VersionString() + "\r\nMode\tIdeas\tWords\tDensity\r\n";
      textBoxDetails.Clear();
      textBoxDetails.Text = VersionString() + "\r\n\r\n";
      // Make sure neither of them is focused.
      // The following setting tends to revert back to
      // the wrong value in Forms Designer.
      textBoxResults.TabStop = false;
      textBoxDetails.TabStop = false;
    }

    void InitializeTagger()  // Initialize the tagger and say what you're doing
    {
      // Disallow clicking on Form1
      this.Enabled = false;

      // Temporarily change the text of textBoxResults *and* 
      // put up a splash screen during this time-consuming operation.
      string saveText = textBoxResults.Text;
      textBoxResults.Text =
        "CPIDR is initializing the tagger.\r\n" +
        "This process can take up to 3 minutes.\r\n" +
        "Please wait...";
      AboutBox1 pop = new AboutBox1("CPIDR is initializing the tagger.  Please wait...");
      pop.Show();

      System.Media.SystemSounds.Exclamation.Play();  // audible indication of action
      Application.UseWaitCursor = true;

      // Initialize the tagger in its own thread so that the main window remains responsive
      Thread init = new Thread(new ThreadStart(Tagger.Initialize));
      init.Start();
      
      // Wait for the thread to finish while still processing events
      // Note: init.Join() is not sufficient here; the main window doesn't get updated enough.
      while (init.IsAlive)
      {
        Application.DoEvents();
      }

      Application.UseWaitCursor = false;
      System.Media.SystemSounds.Exclamation.Play();  // audible indication of readiness

      this.Enabled = true;

      pop.Close();
      //splash.Abort();

      textBoxResults.Text = saveText;
    }

    #endregion


    #region The process of analyzing a sentence or text

    /// <summary>
    /// Corrects some common non-ASCII characters prior to tagging
    /// </summary>
    string ASCIIfy(string s)
    {
      StringBuilder b = new StringBuilder("");

      foreach (char c in s)
      {
        switch (c)
        {
          case '':
          case '':
            b.Append('\'');
            break;

          case '':
          case '':
            b.Append('"');
            break;

          default:
            b.Append(c);
            break;
        }
      }

      return b.ToString();
    }


    void AnalyzeString(string InputString, string LabelToDisplayInResults)
    {
      if (InputString.Trim() == "") return;  // bail out if there is no input
      
      if (!Tagger.Initialized) InitializeTagger();

      Cursor.Current = Cursors.WaitCursor;

      int wcount;
      int pcount;
      double density;
      bool speechMode = checkBoxRejectFillers.Checked;  // parameter for the analysis

      // Perform analysis

      string t = Tagger.TagText(ASCIIfy(InputString));

      IdeaDensityRater.WordList wlist = 
        IdeaDensityRater.RateTaggedText(t, speechMode, out wcount, out pcount);

      density = (float)pcount / wcount;

      textBoxResults.AppendText(
        String.Format("{0}\t{1,5}\t{2,5}\t  {3:0.000}\t\"{4}\"\r\n",
                      (speechMode ? "Spch" : "Norm"),
                      pcount,
                      wcount,
                      density,
                      LabelToDisplayInResults.Replace("\"","\\\"") // escape any quotes that are in the label
                      )
      );

      textBoxDetails.AppendText(
        "\r\n\r\n\"" +
        LabelToDisplayInResults.Replace("\"", "\\\"") + 
        "\"\r\n" +
        (speechMode ? "(Speech Mode)\r\n" : "" ) +
        wlist.ToString() + "\r\n" +
        String.Format(
           " {0,5} propositions\r\n {1,5} words\r\n {2:0.000} density\r\n\r\n",
           pcount, 
           wcount,
           density)
      );

      Cursor.Current = Cursors.Default;

    }

    #endregion

    #region User interface actions - some of which may be called > 1 way
    

    void AnalyzeSentence()
    {
      string label = textBoxSingleSentenceInput.Text;
      if (label.Length > 40) label = label.Substring(0, 37) + "...";
      AnalyzeString(textBoxSingleSentenceInput.Text, label);
      textBoxSingleSentenceInput.Clear();
    }


    void AnalyzeFiles()
    {
      //openFileDialog1.FileName = "";  // do not redisplay previous one if any
      DialogResult r = openFileDialog1.ShowDialog();

      if (r == DialogResult.OK)
      {
        // Take the filenames in alphabetical order
        List<string> names = new List<string>(openFileDialog1.FileNames);
        names.Sort();
        foreach (string s in names)
        {
          AnalyzeOneFile(s);
          Application.DoEvents();
        }
      }
    }


    void AnalyzeOneFile(string filename)
    {
      string shortname = System.IO.Path.GetFileName(filename);
      try
      {
        string text = System.IO.File.ReadAllText(filename);
        AnalyzeString(text, shortname);
      }
      catch (Exception e)
      {
        MessageBox.Show("Unable to process file: " + filename + "\r\n\r\n" + e.Message, "CPIDR");
      }
    }

    
    void SaveTextBoxContents(TextBox t, string defaultFileName, string filter)
    {
      saveFileDialog1.FileName = defaultFileName;
      saveFileDialog1.Filter = filter;
      DialogResult r = saveFileDialog1.ShowDialog();
      if (r == DialogResult.OK)
      {
        try
        {
          System.IO.File.WriteAllText(saveFileDialog1.FileName, t.Text);
        }
        catch (Exception e)
        {
          MessageBox.Show("Unable to write file.\r\n\r\n" + e.Message, "CPIDR");
        }
      }
    }
    
    void SaveResults()
    {
      SaveTextBoxContents(textBoxResults,"CPIDR-results.txt","Tab-delimited text file|*.txt");
    }

    void SaveDetails()
    {
      SaveTextBoxContents(textBoxDetails,"CPIDR-details.txt","Text file|*.txt");
    }

    void ShowAboutBox()
    {
      AboutBox1 ab = new AboutBox1();
      ab.ShowDialog();
    }


    #endregion

    #region Menu bar items

    private void buttonAnalyzeSentence_Click(object sender, EventArgs e)
    {
      AnalyzeSentence();
    }

    private void analyzeSentenceToolStripMenuItem_Click(object sender, EventArgs e)
    {
      AnalyzeSentence();
    }


    private void analyzeFilesToolStripMenuItem_Click(object sender, EventArgs e)
    {
      AnalyzeFiles();
    }

    private void buttonAnalyzeFiles_Click(object sender, EventArgs e)
    {
      AnalyzeFiles();
    }


    private void buttonSaveResults_Click(object sender, EventArgs e)
    {
      SaveResults();
    }

    private void saveResultsToFileToolStripMenuItem_Click(object sender, EventArgs e)
    {
      SaveResults();
    }

    private void buttonSaveDetails_Click(object sender, EventArgs e)
    {
      SaveDetails();
    }

    private void saveDetailsToFileToolStripMenuItem_Click(object sender, EventArgs e)
    {
      SaveDetails();
    }

    private void exitToolStripMenuItem_Click(object sender, EventArgs e)
    {
      Application.Exit();
    }

    private void clearOutputWindowsToolStripMenuItem_Click(object sender, EventArgs e)
    {
      InitializeTextBoxes();
    }

    private void aboutCPIDR3ToolStripMenuItem_Click(object sender, EventArgs e)
    {
      ShowAboutBox();
    }

    #endregion



  }
}