using System;
using System.Collections.Generic;
using System.Text;

namespace CPIDR
{
  /// <summary>
  /// This part of the IdeaDensityRater class contains only the
  /// method WordList.ApplyIdeaCountingRules, i.e., the rule set.
  /// </summary>
  static partial class IdeaDensityRater
  {
    public partial class WordList
    {

      // We use the Penn Treebank tag set as output by MontyTagger.
      // In addition, we introduce the following tags of our own:
      //    NOT     for negatives (not, n't)
      //    blank   for empty positions and for deleted items (items to be ignored)

      #region Predefined sets of tags and tokens

      // A faster kind of searching can almost certainly be implemented here.
      // Right now, this isn't the bottleneck; these are chosen for efficient coding.

      /// <summary>
      /// Tags for punctuation marks
      /// </summary>
      static string[] Punct = { ":", ",", "." };

      /// <summary>
      /// Tags for adjectives
      /// </summary>
      static string[] Adj = { "JJ", "JJR", "JJS" };

      /// <summary>
      /// Tags for adverbs
      /// </summary>
      static string[] Adv = { "RB", "RBR", "RBS", "WRB" };

      /// <summary>
      /// Tags for verb forms.
      /// </summary>
      static string[] Verb = { "VB", "VBD", "VBG", "VBN", "VBP", "VBZ" };

      /// <summary>
      /// Tags for noun forms.
      /// </summary>
      static string[] Noun = { "NN", "NNS", "NNP", "NNPS" };

      /// <summary>
      /// Tags for interrogatives
      /// </summary>
      static string[] Interrogative = { "WDT", "WP", "WPS", "WRB" };

      /// <summary>
      /// All the tags that, by default, are taken to be propositions.
      /// </summary>
      static string[] Prop = 
      {
        "CC",                                       // coordinating conjunction
        "CD",                                       // cardinal numeral
        "DT",                                       // determiner
        "IN",                                       // preposition/subordinating conj.
        "JJ", "JJR", "JJS",                         // adjective
        "PDT",                                      // predeterminer
        "POS",                                      // possessive "'s" (which is not counted as a word)
        "PRP$",                                     // possessive pronoun
        "PP$",                                      // possessive pronoun
        "RB", "RBR", "RBS",                         // adverbs
        "TO",                                       // "to" whether prep. or infinitival
        "VB", "VBD", "VBG", "VBN", "VBP", "VBZ",    // verbs
        "WDT", "WP", "WPS", "WRB"                   // interrogatives/relatives
      };


      /// <summary>
      /// Words that are often non-propositional fillers.
      /// A sentence consisting wholly of these is propositionless.
      /// </summary>
      static string[] Filler =
      { 
        "and", "or", "but", "if", "that", "just", "you", "know"
      };


      /// <summary>
      /// All forms of 'be'.
      /// </summary>
      static string[] Be = 
      { 
        "am", "is", "are", "was", "were", "being", "been" 
      };


      /// <summary>
      /// Common negative contractions that may slip through the tagger, especially
      /// if accidentally typed without the apostrophe.
      /// </summary>
      static string[] Nt =
      {
        "didn't",   "didnt",
        "don't",    "dont",
        "can't",    "cant",
        "couldn't", "couldnt",
        "won't",    "wont",
        "wouldn't", "wouldnt"
      };

      
      /// <summary>
      /// 'Come', 'go', and their synonyms form a single proposition with following 'to' or 'from'.
      /// </summary>
      static string[] ComeGo = 
      {
        "come",    "comes",    "came",                 "coming",
        "return",  "returns",  "returned",             "returning",
        "arrive",  "arrives",  "arrived",              "arriving",
        "go",      "goes",     "went",      "gone",    "going",
        "depart",  "departs",  "departed",             "departing",
        "emanate", "emanates", "emanated",             "emanating" 
      };

      /// <summary>
      /// All forms of all auxiliary verbs.
      /// </summary>
      static string[] Aux = 
      { 
        "be",   "am", "is", "are", "was", "were", "being", "been",
        "have", "has",             "had",         "having",
        "do",   "does",            "did",         // "doing" and "done" are not aux forms
        "need",         // "needs" is not an aux
        "dare"          // "dares" is not an aux
        // and we need not include modals here
        // because the tagger does not tag them as verbs.
      };

      /// <summary>
      /// Linking verbs:
      /// all forms of all verbs that take an adjective after them.  
      /// </summary>
      static string[] Link = 
      { 
        // Being
        "be",     "am", "is", "are",   "was", "were",   "been",     "being", 
        // Becoming
        "become", "becomes",           "became",                    "becoming",
        "get",    "gets",              "got",           "gotten",   "getting",
        // Seeming visually
        "look",   "looks",             "looked",                    "looking",
        "seem",   "seems",             "seemed",                    "seeming",
        "appear", "appears",           "appeared",                  "appearing",
        // Seeming through other senses
        "sound",   "sounds",           "sounded",                   "sounding",
        "feel",    "feels",            "felt",                      "feeling",
        "smell",   "smells",           "smelled",                   "smelling",
        "taste",   "tastes",           "tasted",                    "tasting"
      };

      /// <summary>
      /// Causative linking verbs:
      /// all forms of all verbs that take noun phrase + adjective after them,
      /// such as "make it better" or "turn it green."
      /// </summary>
      static string[] CLink =
      {
        //  What else needs to be included here?
        "make",   "makes",   "made",    "making",
        "turn",   "turns",   "turned",  "turning",
        "paint",  "paints",  "painted", "painting"
      };

      /// <summary>
      /// First elements of correlating conjunctions
      /// </summary>
      static string[] Correl = 
      { 
        "both", "either", "neither" 
      };

      // N.B. The following are not all the negative-polarity items of English,
      // but only the ones that seem to form 2-word 1-concept idioms.

      /// <summary>
      /// Negative-polarity items where the negative, rather than this word,
      /// counts as the proposition; e.g., "not...yet" = "not".
      /// </summary>
      static string[] NegPol1 = 
      { 
        "yet", "much", "many", "any", "anymore"
      };

      /// <summary>
      /// Negative-polarity items where this word, rather than the earlier 
      /// negative, counts as the proposition; e.g., "not...unless" = "(n)unless".
      /// </summary>
      static string[] NegPol2 = 
      { 
        "unless"  // are there others?
      }; 


      #endregion

      #region Utility methods

      /// <summary>
      /// Tests whether an array of strings contains an item.
      /// (A faster search is surely possible.)
      /// </summary>
      public bool Contains(string[] set, string item)
      {
        return (Array.IndexOf(set, item) > -1);
      }

      /// <summary>
      /// Finds the first word in the sentence containing w[i]
      /// (which may be w[i] itself) and returns its index.
      /// Assumes valid input.
      /// </summary>
      /// <returns></returns>
      public int BeginningOfSentence(List<WordListItem> w, int i)
      {
        int j = i - 1;
        while ((j > 0) && (w[j].tag != ".") && (w[j].tag != "")) j--;
        return j+1;
      }

      /// <summary>
      /// Determines whether a word is likely to be a repetition of another.
      /// The first word may be incomplete, e.g. "hesi- hesitation".
      /// </summary>
      bool Match(string first, string second)
      {
        if (first.Length == 0 || second.Length == 0) return false;
        if (first == second) return true;
        if (first.EndsWith("-")) first = first.Substring(0, first.Length - 1);  // final hyphen drop
        if ((second.Length > 3)
          && (first != "a")
          && (first != "an")
          && second.StartsWith(first)) return true;
        return false;
      }

      #endregion

      
      /// <summary>
      /// Applies all the idea-counting rules to a WordList.
      /// </summary>
      public void ApplyIdeaCountingRules(bool speechMode)
      {
        // The rules are numbered.  The most recent rule to act on each item
        // is identified in its .rulenumber field.

        // Rule numbering is not consecutive so that there can be room to
        // add others in each group.

        // Some rules apply only if speechMode is true.
        // That is an option that can be set when analyzing transcribed speech
        // that has not been edited.  It may result in undercounting of
        // well-edited English.

        List<WordListItem> b = this.items;  
        // This enables us to refer to "b" rather than "this.items" in the rules.
        // b is a pointer; that is, b and this.items are the same object.

        for (int i=0; i<b.Count; i++)
        //
        // This loop iterates through the WordList and may add and remove items.
        //
        // The rules look back toward the beginning from the current word; that is,
        // the rule is triggered by the last word in the pattern that it is looking for.
        // There are guaranteed to be 10 null items at the beginning of the WordList,
        // so rules do not have to worry about going past the beginning of the list.
        //
        // Most rules look at the output of prior rules; rule 200 is a good example
        // of this.  The program could be speeded up by identifying rules that do *not*
        // feed a subsequent rule, and putting a "continue" statement in them, as in
        // rule 000.
        //
        // When an addition or deletion to the WordList is to be made, it must take 
        // place AFTER the current location (e.g., when looking at b[i] you can add 
        // or delete at i+1 but not at i-1), to prevent renumbering.  Rules 003 and 004
        // contain examples of stepping backward and deleting forward.
        //
        // Remember that a deletion removes the item from the word count as well as
        // the idea count.  Accordingly, items should only be deleted if they've
        // been moved or should not be counted as words.
        //
        {
          // Debugging trace (executes only in debug mode)
          System.Diagnostics.Debug.WriteLine(i + " " + b[i].token + "/" + b[i].tag);

          // Rule group 000 - Identify words and adjust tags

          // (Some of these rules are very specific for common
          // errors in the output of MontyTagger.)

          // 000
          // If it's a null item, skip the rest of these tests.
          // There will always be at least 10 null items at the beginning
          // so we can freely look back from the current position
          // without running off the beginning of the list.
          if (b[i].token == "") continue;

          // 001
          // The symbol  ^  used to mark broken-off spoken sentences
          // is an end-of-sentence marker.
          if (b[i].token == "^")
          {
            b[i].tag = ".";
            b[i].rulenumber = 1;
          }

          // 002
          // The item is a word if its token starts with a letter or digit
          // and its tag is not SYM (symbol).
          if (Char.IsLetterOrDigit(b[i].token[0]) && (b[i].tag != "SYM"))
          {
            b[i].isword = true;
            b[i].rulenumber = 2;
          }

          // 003
          // Two cardinal numbers in immediate succession are combined into one
          // (Uncommon situation; the next one is much more common)
          if ((b[i].tag == "CD") && (b[i - 1].tag == "CD"))
          {
            b[i - 1].token = b[i - 1].token + " " + b[i].token;  // adjust token
            b[i - 1].rulenumber = 3;
            i--;                                                 // step backward 1
            b.RemoveAt(i + 1);                                   // delete forward 1
          }

          // 004
          // Cardinal + nonalphanumeric + cardinal are combined into one token
          // (Common situation, for handling fractions, decimals, etc.
          if (
              (b[i].tag == "CD")    // current token is a number
            && 
              (b[i-1].token.Length > 0)  // preceding token contains some characters
            &&
              !Char.IsLetterOrDigit(b[i - 1].token[0]) // the first of which is nonalphanumeric
            && 
              (b[i - 2].tag == "CD") // pre-preceding token is also a number
             )
          {
            b[i - 2].token = b[i - 2].token + b[i - 1].token + b[i].token;  // adjust token
            b[i - 2].rulenumber = 4;
            i = i - 2;          // step backward 2
            b.RemoveAt(i + 1);  // delete forward 2
            b.RemoveAt(i + 1);
          }

          // 020
          // Repetition of the form "A A" is simplified to "A".
          // The first A can be an initial substring of the second one.
          // Both remain in the word count. 
          if (speechMode)
          {
            if (Match(b[i - 1].token, b[i].token))
            {
              // Mark the first A as to be ignored
              b[i - 1].isprop = false;
              b[i - 1].isword = false;
              b[i - 1].tag = "";
              b[i - 1].rulenumber = 20;
            }
          }

          // 021, 022
          // Repetition of the form "A Punct A" is simplified to "A".
          // Repetition of the form "A B A" is simplified to "A B".
          // Both A's remain in the word count.
          // The first A can be an initial substring of the second one.
          // Punct is anything with tag "." or "," or ":".
          if (speechMode)
          {
            if (Match(b[i - 2].token, b[i].token) && !Contains(Punct, b[i].tag))
            {
              // Mark the first A to be ignored
              b[i - 2].tag = "";
              b[i - 2].isword = false;
              b[i - 2].isprop = false;
              b[i - 2].rulenumber = 22;
              // Mark the punctuation mark to be ignored
              if (Contains(Punct, b[i - 1].tag))   // Punct
              {
                b[i - 1].tag = "";
                b[i - 1].isword = false;
                b[i - 1].isprop = false;
                b[i - 1].rulenumber = 21;
              }
            }
          }

          // 023
          // Repetition of the form "A B Punct A B" is simplified to "A B".
          // Both A's and B's remain in the word count.
          // The first A (or B) can be an initial substring of the second one.
          // Punct is anything with tag "." or "," or ":".
          if (speechMode)
          {
            if (Match(b[i - 3].token, b[i].token)
                && Match(b[i - 4].token, b[i - 1].token)
                && ((b[i - 2].tag == ".") || (b[i - 2].tag == ",") || (b[i - 2].tag == ":"))
              )
            {
              b[i - 4].tag = b[i - 3].tag = b[i - 2].tag = "";
              b[i - 4].isword = b[i - 3].isword = b[i - 2].isword = false;
              b[i - 4].isprop = b[i - 3].isprop = b[i - 2].isprop = false;
              b[i - 4].rulenumber = b[i - 3].rulenumber = b[i - 2].rulenumber = 23;
            }
          }
    
          // 050
          // 'not' and any word ending in "n't" are putatively propositions and 
          // their tag is changed to NOT.
          if ((b[i].token == "not") || (b[i].token.EndsWith("n't")) || Contains(Nt,b[i].token))
          {
            b[i].isprop = true;
            b[i].tag = "NOT";
            b[i].rulenumber = 50;
          }

          // 054
          // 'that/DT' or 'this/DT' is a pronoun, not a determiner, if the word following it
          // is any kind of verb or adverb.
          if (   ((b[i - 1].token == "that") || (b[i-1].token == "this")) 
               && 
                 (Contains(Verb, b[i].tag) || Contains(Adv,b[i].tag))
             )
          {
            b[i - 1].tag = "PRP";
            b[i - 1].rulenumber = 54;
            b[i - 1].isprop = false;
          }


          // Rule group 100 - Word order adjustment
          // Since all the rules apply in one pass,
          // this must be approached carefully.

          // 101
          // Subject-Aux inversion
          // If the current word is an Aux,
          // and the current word is the first word of the sentence
          // or the sentence begins with an interrogative,
          // move the current word rightward to put it in front 
          // of the first verb, or the end of the sentence.
          // In some cases this will move a word too far to the right,
          // but the effect on proposition counting is benign.
          if (Contains(Aux,b[i].token))
          {
            int bos = BeginningOfSentence(b,i);
            if ((bos == i) || Contains(Interrogative,b[bos].tag))
            {
              // find out where to move to
              int dest;
              for (dest = i + 1; dest < b.Count; dest++)
              {
                if ((b[dest].tag == ".") || Contains(Verb, b[dest].tag)) break;
              }
              // if movement is called for,
              if (dest > i + 1)         
              {
                // insert a copy in the new location
                b.Insert(dest, new WordListItem(b[i].token, b[i].tag, true, true, 101));
                // mark the old item as to be ignored
                b[i].tag = "";
                b[i].isprop = false;
                b[i].isword = false;
                b[i].token += "/moved"; 
              }
            }
          }


          // Rule group 200 - Preliminary identification of propositions

          // 200
          // The tags in Prop are taken to indicate propositions.
          if (Contains(Prop,b[i].tag))
          {
            b[i].isprop = true;
            b[i].rulenumber = 200;
          }

          // 201
          // 'The', 'a', and 'an' are not propositions.
          if ((b[i].token == "the") || (b[i].token == "an") || (b[i].token == "a"))
          {
            b[i].isprop = false;
            b[i].rulenumber = 201;
          }


          // We get better agreement with Turner & Greene without this rule.
          //// 202
          //// An attributive noun (such as 'lion' in 'lion tamer') is a proposition
          //// (like an adjective).
          //if (Contains(Noun, b[i].tag) && (b[i - 1].tag == "NN"))
          //{
          //  b[i - 1].isprop = true;
          //  b[i - 1].rulenumber = 202;
          //}


          // 203
          // The first word in a correlating conjunction such as
          // "either...or", "neither...nor", "both...and"
          // is not a proposition. The second word is tagged CC;
          // the first word may have been tagged CC or DT.
          if ((b[i].tag == "CC") && (!Contains(Correl, b[i].token)))
          {
            // Search back up to 10 words, but not across a sentence end.
            for (int j = i - 1; j > i - 10; j--)
            {
              if (b[j].tag == ".") break;
              if (Contains(Correl, b[j].token))
              {
                b[j].isprop = false;
                b[j].rulenumber = 203;
                break;
              }
            }
          }


          // 204
          // "And then" and "or else" are each a single proposition
          if ((b[i - 1].token == "and") && (b[i].token == "then")
               || (b[i - 1].token == "or") && (b[i].token == "else")
             )
          {
            b[i].isprop = false;
            b[i].rulenumber = 204;
          }


          // 206
          // "To" is not a proposition when it is last word in sentence.
          if ((b[i].tag == ".") && (b[i - 1].tag == "TO"))
          {
            b[i - 1].isprop = false;
            b[i - 1].rulenumber = 206;
          }

          // 207
          // Modal is a proposition when it is last word in sentence.
          if ((b[i].tag == ".") && (b[i - 1].tag == "MD"))
          {
            b[i - 1].isprop = true;
            b[i - 1].rulenumber = 207;
          }
          
          // 210
          // Cardinal number is a proposition only if there is a noun
          // within 5 words after it (not crossing a sentence boundary).
          // This is so "in 3 parts" is 2 props but "in 1941" is only one.
          if (b[i].tag == "CD")
          {
            b[i].isprop = false;
            b[i].rulenumber = 210;
            // This rule looks forward up to 5 words, but not past
            // an end-of-sentence marker or the end of the WordList
            for (int j = i + 1; (j < b.Count) && (j < i + 6); j++)
            {
              if (b[j].tag == ".") break;
              if (Contains(Noun,b[j].tag))
              {
                b[i].isprop = true;
                break;
              }
            }
          }

          // 211
          // 'Not...unless' and similar pairs count as one proposition
          // (the second word is the one counted).
          if (Contains(NegPol2,b[i].token))
          {
            // Much the same algorithm as for correlating conjunctions.
            // Search back up to 10 words, but not across a sentence end.
            for (int j = i - 1; j > i - 10; j--)
            {
              if (b[j].tag == ".") break;
              if (b[j].tag == "NOT")
              {
                b[j].isprop = false;
                b[j].rulenumber = 211;
                break;
              }
            }
          }

          // 212
          // 'Not...any' and similar pairs count as one proposition
          // (the first word is the one counted).
          if (Contains(NegPol1, b[i].token))
          {
            // Much the same algorithm as for correlating conjunctions.
            // Search back up to 10 words, but not across a sentence end.
            for (int j = i - 1; j > i - 10; j--)
            {
              if (b[j].tag == ".") break;
              if (b[j].tag == "NOT")
              {
                b[i].isprop = false;
                b[i].rulenumber = 212;
                break;
              }
            }
          }

          // 213
          // "Going to" is not a proposition when immediately preceding a verb.
          if (Contains(Verb,b[i].tag) && (b[i-1].token == "to") && (b[i-2].token == "going"))
          {
            b[i-1].isprop = false;
            b[i-1].rulenumber = 213;
            b[i-2].isprop = false;
            b[i-2].rulenumber = 213;
          }

          // 214
          // "If ... then" is 1 conjunction, not two.
          // Actually checking for "if ... then (word)"
          // because "then" as last word of sentence is more likely to be adverb.
          if (b[i].isword && (b[i-1].token == "then"))
          {
            // Much the same algorithm as for correlating conjunctions.
            // Search back up to 10 words, but not across a sentence end.
            for (int j = i - 1; j > i - 10; j--)
            {
              if (b[j].tag == ".") break;
              if (b[j].token == "if")
              {
                b[i-1].isprop = false;
                b[i-1].rulenumber = 214;
                break;
              }
            }
          }

          // 225
          // "each other" is a pronoun (to be tagged as PRP PRP).
          if ((b[i].token == "other") && (b[i - 1].token == "each"))
          {
            b[i].tag = b[i - 1].tag = "PRP";
            b[i].isprop = b[i - 1].isprop = false;
            b[i].rulenumber = b[i - 1].rulenumber = 225;
          }


          // 230
          // "how come" and "how many" are each 1 proposition, not two.
          if (
                ((b[i].token == "come") || (b[i].token == "many"))
              &&
                (b[i - 1].token == "how")
            )
          {
            b[i].isprop = false;
            b[i].tag = b[i - 1].tag;
            b[i].rulenumber = 230;
          }


          // Rule group 300 - Linking verbs

          // 301
          // Linking verb is not a proposition if followed by adj. or adv.
          // (Apparently, adverbs are frequent tagging mistakes for adjectives.)
          if ((Contains(Adj,b[i].tag) || Contains(Adv,b[i].tag)) && Contains(Link, b[i - 1].token))
          {
            b[i - 1].isprop = false;
            b[i - 1].rulenumber = 301;
          }

          // 302
          // "Be" is not a proposition when followed by a preposition.
          // (May want to modify this to allow an intervening adverb.)
          if ((b[i].tag == "IN") && Contains(Be,b[i-1].token))
          {
            b[i - 1].isprop = false;
            b[i - 1].rulenumber = 302;
          }

          // 310
          // Linking verb + Adverb + { PDT, DT } is 2 propositions
          // (e.g., "he is now the president").
          // (Would otherwise be undercounted because of rule 201).
          if ( (b[i].tag == "DT") || (b[i].tag == "PDT") )
          {
            if (Contains(Adv,b[i - 1].tag) && Contains(Link, b[i - 2].token))
            {
              b[i - 1].isprop = true;
              b[i - 1].rulenumber = 310;
              b[i - 2].isprop = true;
              b[i - 2].rulenumber = 310;
            }
          }

          // 311
          // Causative linking verbs: 'make it better' and similar
          // phrases do not count the adjective as a new proposition
          // (since the verb was counted).
          if (Contains(Adj,b[i].tag))
          {
            // Much the same algorithm as for correlating conjunctions.
            // Search back up to 10 words, but not across a sentence end.
            for (int j = i - 1; j > i - 10; j--)
            {
              if (b[j].tag == ".") break;
              if (Contains(CLink,b[j].token))
              {
                b[i].isprop = false;
                b[i].rulenumber = 311;
                break;
              }
            }
          }


          // Rule group 400 - Auxiliary verbs are not propositions

          // Note that Verb is a set of tags but Aux is a set of tokens.
          // Note also that Verb includes all Aux.

          // 401
          // Aux not is one proposition, not two
          if ((b[i].token == "not") && Contains(Aux,b[i-1].token))
          {
            b[i - 1].isprop = false;
            b[i - 1].rulenumber = 401;
          }

          // 402
          // Aux Verb is one proposition, not two
          if (Contains(Verb,b[i].tag) && Contains(Aux,b[i-1].token))
          {
            b[i - 1].isprop = false;
            b[i - 1].rulenumber = 402;
          }

          // 405
          // Aux NOT Verb          (NOT and the second Verb are propositions)
          // Also Aux Adverb Verb  (e.g., "had always sung", "would rather go")
          if (Contains(Verb,b[i].tag) && 
                ( (b[i-1].tag == "NOT") || Contains(Adv,b[i-1].tag) ) && Contains(Aux,b[i-2].token)
             )
          {
            b[i - 2].isprop = false;
            b[i - 2].rulenumber = 405;
          }


          // Rule group 500 - Constructions involving 'to'

          // 510
          // TO VB is one proposition, not two
          if ((b[i].tag == "VB") && (b[i-1].tag == "TO"))
          {
            b[i - 1].isprop = false;
            b[i - 1].rulenumber = 510;
          }

          // 511
          // "for ... TO VB": "for" is not a proposition
          if ((b[i].tag == "VB") && (b[i - 1].tag == "TO"))
          {
            // Search back up to 10 words, but not across a sentence end.
            for (int j = i - 1; j > i - 10; j--)
            {
              if (b[j].tag == ".") break;
              if (b[j].token == "for")
              {
                b[j].isprop = false;
                b[j].rulenumber = 511;
                break;
              }
            }
          }

          //// 512
          //// "From" and "to" form a single proposition with
          //// preceding "go", "come", or their synonyms.
          //// (We get better agreement with Turner & Greene without this rule.)
          //if (
          //      ( (b[i].token == "to") || (b[i].token == "from") )
          //    &&
          //      (Contains(ComeGo, b[i - 1].token) || Contains(ComeGo, b[i - 2].token))
          //   )
          //{
          //  b[i].isprop = false;
          //  b[i].rulenumber = 512;
          //}


          // Rule group 600 - Fillers

          // 610
          // A sentence consisting entirely of probable filler words is propositionless
          if (speechMode && b[i].tag == ".")
          {
            int bos = BeginningOfSentence(b, i);
            int k = 0;
            for (int j = bos; j < i; j++)
            {
              if ((b[j].tag != "UH") && !Contains(Filler, b[j].token)) k++;
            }
            if (k == 0)
            {
              for (int j = bos; j < i; j++)
              {
                b[j].tag = "";
                b[j].isprop = false;
                b[j].rulenumber = 610;
              }
            }
          }

          // 632
          // In speech mode, "like" is a filler when not immediately preceded by a form of "be".
          if (speechMode)
          {
            if ((b[i].token == "like") && !Contains(Be, b[i - 1].token))
            {
              b[i].tag = "";
              b[i].isprop = false;
              b[i].rulenumber = 632;
            }
          }

          // 634
          // In speech mode, "you know" is a filler and counts as one word, not two.
          if (speechMode)
          {
            if ((b[i - 1].token == "you") && (b[i].token == "know"))
            {
              // back up one
              i--;
              // delete forward one
              b.RemoveAt(i + 1);
              // reset data for the current item
              b[i].token = "you_know";
              b[i].tag = "";
              b[i].isprop = false;
              b[i].isword = true;
              b[i].rulenumber = 634;
            }
          }



        } // end for

      } // end ApplyIdeaCountingRules()


    }
  }
}